/******************************************************************************
NeuralNetLN filter plugin for Avisynth by V.C.Mohan
NeuralNetLN filter trains on input clip startframe and trainer clip tframe, then 
processes input frames. Uses Resilient propogation. Linear type.
Author V.C.Mohan
Nov 21, 2005
********************************************************************************/  
//#include "windows.h"
//#include "avisynth.h"
//#include <cmath>

		



class NeuralNetLN : public GenericVideoFilter {
		PClip trainer;
		int nl;		//start frame number of  (child) clip. This and tframe will be used for training
		int enl;	// end frame number for this line
		int tframe;	// frame number of trainer clip to use for training
		 
	
		int slx;	// start x coordinate of rectangle left 
		int sty;	// start y coordinate of rectangle top  
		int srx;	// start x coord of window right 
		int sby;	// start y coord of window bottom 
		int elx;	// end x coordinate of rectangle left 
		int ety;	// end y coordinate of rectangle top  
		int erx;	// end x coord of window right 
		int eby;	// end y coord	of window bottom 
		int xpts;	// number of points along x axis to be used.. Only odd numbers 1 to 121
		int ypts;	// number of points along y axis to be used. Only odd numbers 1 to 121
		int tlx, tty, trx, tby;	// trainer clip window coordinates
		double hf;		// hidden layer weight factor
		int iter;	// number of iterations for training
		
		int wset;	// wset  weights are this set (random numbers start after skipping wset-1 * total weights) 
		int best;		// use best weights out of this number of different starting weights
		bool line;			// training window line or rectangle
	
		unsigned char *input;	// buffer to hold input values each time plus one bias
		double  *wh, *whcopy, *dedwt1, *dedwt2;	// weights for each node( in cluding bias) of input layer and de/dwt values, 
		bool exist;			// flag for creation by new
		double *errbuf;
		bool trainingOver;
		int inodes;	//	(xpts*ypts+1)
		double *deltaw, *delta;
		
		double deltamax, deltamin;
		double yetaplus, yetaminus;
		double emax;
  public:
							//Definition of function
    NeuralNetLN(PClip _child,PClip _trainer,
				int _nl,int _enl,
				
				int _slx,int _sty,int _srx, int _sby,
				int _elx,int _ety,int _erx, int _eby,
				
				int _xpts, int _ypts,
				int _tlx, int _tty, int _trx, int _tby,
				int _tframe,
				double _hf,  int _iter,
				 
				int _wset, 
				
				int _best,
				
				bool _line,
				IScriptEnvironment* env) ;	
			
	~NeuralNetLN();				//destructor

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
	void  getcase(const unsigned char * fp, int fpitch,int kb, int h,int w);
	double	getoutput(void);
	int sign (double val);
	void adjustWeights(double *dwold, double *dwnew);
	void sumdedwt(double err, double *dwtnew);

};	
/*************************************************
 * The following is the implementation 
 * of the defined functions.
 *************************************************/
//Here is the acutal constructor code used

NeuralNetLN::NeuralNetLN(PClip _child,PClip _trainer,
			    int _nl,int _enl,
				
				int _slx,int _sty,int _srx, int _sby,
				int _elx,int _ety,int _erx, int _eby,
				
				int _xpts, int _ypts,
				int _tlx, int _tty, int _trx, int _tby,
				int _tframe,
				double _hf,  int _iter,
				 
				int _wset,int _best,
				
				bool _line,
				IScriptEnvironment* env) :

	GenericVideoFilter(_child),trainer(_trainer),
						nl(_nl),enl(_enl),
			
						slx(_slx), sty(_sty),
						srx(_srx), sby(_sby),
						elx(_elx), ety(_ety),
						erx(_erx), eby(_eby),
						
						xpts(_xpts), ypts(_ypts),
						tlx(_tlx), tty(_tty),
						trx(_trx), tby(_tby),
						
						tframe(_tframe),
						hf(_hf),  iter(_iter),
						
						wset(_wset), 
						
						best(_best),
						
						line (_line)

						
					
						
						 
			
{
	
	exist = false;
	trainingOver = false;
	inodes= xpts*ypts+1;
	if(hf>0)
	{
		deltamax=5.0 * hf;
		deltamin=exp(-6)*hf;
	}
	else
	{
		deltamax=5.0 * 0.001;
		deltamin=exp(-6)*0.001;
	}

	yetaplus = 1.2;
	yetaminus = 0.5;
	emax=0;

}

 NeuralNetLN::~NeuralNetLN()
{
	delete []wh;	// hidden layer weights
	delete []errbuf;
	delete []input;
	 if(exist)
	 {
		 
		 
		 delete []whcopy;
		 delete []deltaw;
		 delete []delta;
		 
		 delete []dedwt1;
		 delete []dedwt2;
		 
	 }

	
}

/***************************************************************/	
	PVideoFrame __stdcall NeuralNetLN::GetFrame(int in, IScriptEnvironment* env)
	{

		PVideoFrame Frame;
		int nframes= enl-nl;	// actual intervals
		if (nframes==0)
			nframes++;
		
	
			Frame = child->GetFrame(in, env);
			if(in<nl ||in>enl )
			
				return Frame;
			
			
			int n=in-nl;
		
		
		

		const unsigned char *fp= Frame->GetReadPtr();
		
		const	int bwd = vi.width;
        const	int bht = Frame->GetHeight();		
		const	int fpitch = Frame->GetPitch();
		int kb=vi.BitsPerPixel()/8;

		PVideoFrame work = env->NewVideoFrame(vi);
		unsigned char* wp =work->GetWritePtr();
		const	int wpitch =work->GetPitch();

		


				// calculate parameters for this frame
		
		int lx=slx+((elx-slx)*n)/nframes;
		int ty= sty+((ety-sty)*n)/nframes;
		int by= sby+((eby-sby)*n)/nframes;
		int rx= srx+((erx-srx)*n)/nframes;
		
		
		

		if(vi.IsRGB())
		{
			// so that same code can be used for all color formats
			by=bht-1-(sty+((ety-sty)*n)/nframes);
			ty=bht-1-(sby+((eby-sby)*n)/nframes);
			int temp=bht-1-tby;
			tby = bht-1-tty;
			tty= temp;

		}
		
		int iterscale= iter>bwd-10  ? iter/(bwd-10) + 1 : 1;
		

						;
		// if we are coming here first time and not by reversing on video display button
		if( ! trainingOver)
		{
			PVideoFrame Frame_Start = n==0? Frame : child->GetFrame(nl, env);
			const unsigned char *fp_s= Frame_Start->GetReadPtr();
			
			
				// training is required
			PVideoFrame trframe = trainer->GetFrame(tframe, env);
			const unsigned char *tp= trframe->GetReadPtr();
			const	int tpitch = trframe->GetPitch();
			int xcases=(trx-tlx-xpts-1);
			int ycases=(tby-tty-ypts-1);
			int cases= xcases*ycases;
			// maximum error possible. this is for scaling purpose

			if( !exist)
			{
				wh= new double[inodes];		// buffer to hold hidden layer weights plus  bias
				
				whcopy= new double[inodes];		// buffer to hold best hidden layer weights plus  bias
				dedwt1 = new double [inodes];
				dedwt2 = new double [inodes];
				deltaw = new double [inodes];
				delta = new double [inodes];
				errbuf = new double[iter/iterscale+1];
				input = new unsigned char[inodes];	// buffer to hold input plus one bias node
				
				exist=true;
			}

				// skip (wset-1)* total number weights random numbers
			for(int i=0; i<inodes*(wset-1); i++)
				rand();
				// setup for line input
			int number;
			double xinc, yinc;

			if(line)
			{
				if(abs(tby- tty) > abs(trx-tlx))					
					number = abs(tby-tty )+1;	// number of patterns  for training
				else
					number = abs(trx-tlx )+1;

				xinc = (trx-tlx)/(double)number;
				yinc = (tby-tty)/(double) number;
					
			}
				// repeat training with different start weights "best" of times	
			for(int b=0; b<best;b++)
			{
				
			
				for( i=0;i<inodes; i++)
				{
					// input layer weights between -0.5hf and 0.5hf
					wh[i]= (((double)rand())/RAND_MAX -0.5)*hf;
					// initialize buffers
					delta[i] = deltamin;
					deltaw[i]= 0;
					dedwt1[i] =0;
				}

				emax=0;
				// for toggling buffers
				double *dedwtnew, *dedwtold;
				// now calculate output and error

				for(i=0;i<iter;i++)		
				{
					if( (i & 1) ==0)
					{
						// switch buffers
						dedwtold = dedwt1;
						dedwtnew = dedwt2;
					}
					else
					{
						// switch buffers
						dedwtold = dedwt2;
						dedwtnew = dedwt1;
					}

					double esum=0.0;
					double minesum;
					
					// zero de/dw buffer
					for(int nodes=0; nodes<inodes; nodes++)
						dedwtnew[nodes]=0;
					
					if(line)
					{
						for(int num=0; num<number;num++)
						{
							int h= tty+num*yinc;
							int w = tlx+num * xinc;
								// get next case for training
							getcase(fp_s, fpitch, kb, h,w);		// get next case input data + bias in input buffer
								// get output of the network with current weights after passing thru sigmoid filter
						
							double output=getoutput();
							
							double error=*(tp+(h+ypts/2)*tpitch+kb*(w+xpts/2))-output;
							esum=esum+error*error;
							sumdedwt(error, dedwtnew);
						}	
								// adjust weights so that output value comes  closer to desired value
						

					}
					else
					{

						for(int h=tty; h< tby-ypts+1; h++)
							for(int w= tlx; w<trx-xpts+1; w++)
							{
							

							// get next case for training
								getcase(fp_s, fpitch, kb, h,w);		// get next case input data + bias in input buffer
							
								double output=getoutput();	
								double error=*(tp+(h+ypts/2)*tpitch+kb*(w+xpts/2))-output;
								esum=esum+error*error;
								sumdedwt(error, dedwtnew); 
							}
								// adjust weights so that output value comes  closer to desired value
							
							
							 
					}
					
					adjustWeights(dedwtold, dedwtnew);
					
					emax= emax < esum? esum:emax;
							// to keep track of lowest esum acheived
					if(i==0 && b==0)
					
						minesum=esum;
							// if so far this is the best copy the weights for future use.
					if(minesum > esum)
					{
						for(int nodes=0; nodes<inodes; nodes++)
							whcopy[nodes]=wh[nodes];
						
						minesum=esum;
					}

						// error sum using the final random weights are stored. //
						//Note this may not be least. Otherwise require so many buffers
					if(b== best-1 && (i % iterscale)==0)
						errbuf[i/iterscale]= esum;


				}	// iterations
			
			}	// best of

				// copy weights got for lowest esum

			for(int nodes=0; nodes<inodes; nodes++)
				wh[nodes]=whcopy[nodes];
			
			if(exist)
			{
				delete []whcopy;
				delete []deltaw;
				delete []delta;
		 
				delete []dedwt1;
				delete []dedwt2;
				exist=false;
			}
			

			trainingOver = true;
		
		}	// training end

		
		// as first frame output   training error plots

		if(n==0)
		{
			unsigned char * wpp= wp;
			// zero out output frame for drawing error and normalized histogram
			for(int h=0; h<bht; h++)
			{
				for(int w=0; w<bwd; w++)				
					wpp[kb*w] = 0;
				wpp+=wpitch;
			}

			PVideoFrame work1 = env->NewVideoFrame(vi);
			unsigned char* wp1 =work1->GetWritePtr();
			const	int wpitch1 =work1->GetPitch();

			int vscale = bht>100? 1 : 100 /bht;
		
				
			// plot the error per iteration
			for(int i=0; i< iter/iterscale; i++)
			{
				int er= 100*errbuf[i] / emax; 
				wp[(bht-1-er/vscale)*wpitch + kb*(10+i)]=255;
			}
				// plot vertical scale

			for(int nn=0; nn<11;nn++)
				for(int sht=0;sht<10;sht++)
				{
					wp[(bht-1-(nn*10)/vscale)*wpitch + kb*sht]=128;
					wp[(bht-1-(nn*10)/vscale)*wpitch + kb*(bwd-1-sht)]=128;
				}
				// plot iter scale

			for(nn=0; nn<iter/iterscale; nn+=(iter/iterscale)/10)
				for(int sht=0;sht<5;sht++)
				{
				
					wp[(bht-1-sht)*wpitch+kb*(10+nn)]=128;
					
				}

				// plot training window/ line

			
				

			if(line)
			{
				int number;
				double xinc, yinc;
				if(abs(tby- tty) > abs(trx-tlx))					
					number = abs(tby-tty )+1;	// number of patterns  for training
				else
					number = abs(trx-tlx )+1;

				xinc = (trx-tlx)/(double)number;
				yinc = (tby-tty)/(double) number;

				for(int num =0; num<number; num++)
					*(wp+ (tty+(int)(num*yinc))*wpitch + kb*(tlx+(int)(num*xinc)))=128;
					
			}

			else
			{
				for(int h=tty; h<tby; h++)
				{
					*(wp+h*wpitch + kb* tlx)=128;
					*(wp+h*wpitch + kb* trx)=128;
				}

				for(int w=tlx; w<trx; w++)
				{
					*(wp+tty*wpitch + kb* w)=128;
					*(wp+tby*wpitch + kb* w)=128;
				}
			}



							
		}
		else
		{
			
			// process regular

		
				for(int h=ty; h< by-ypts+1; h++)
					for(int w= lx; w<rx-xpts+1; w++)
					{
						getcase(fp, fpitch, kb, h,w);		// get input data + bias in input buffer
						double output=getoutput();	
							
						if(output>255)
							output=255;
						else
							if(output<0)
								output=0;

						*(wp+(h+ypts/2)*wpitch+kb*(w+xpts/2))=output;

				
					}
			

		}


			if(n==0)
			{
				//  grey  u and v channels
				if(vi.IsYUY2())
				{
					for(int h=ty;h<by;h++)
						for(int w=lx;w<rx;w++)						
							*(wp+h*wpitch+kb*w+1)=128;
				}

				else if(vi.IsYV12())
				{
				
					unsigned char * wpU=work->GetWritePtr(PLANAR_U);
					unsigned char * wpV=work->GetWritePtr(PLANAR_V);
					const int  wpitchUV=work->GetPitch(PLANAR_U);

					for(int h=ty/2;h<by/2;h++)
						for(int w=lx/2;w<rx/2;w++)
						{
						// grey u and v channels
							*(wpU+h*wpitchUV+w)=128;
							*(wpV+h*wpitchUV+w)=128;
						}
				}

				else if(vi.IsRGB())
				{
					for(int h=ty;h<by;h++)
						for(int w=lx;w<rx;w++)
						{
								// grey by making  red green  blue equal .
							*(wp+h*wpitch+kb*w+1)=*(wp+h*wpitch+kb*w+2)=*(wp+h*wpitch+kb*w);
						}
				}
			}

			else
			{
				//  copy  u and v channels
				if(vi.IsYUY2())
				{
					for(int h=ty;h<by;h++)
						for(int w=lx;w<rx;w++)						
							*(wp+h*wpitch+kb*w+1)=*(fp+h*fpitch+kb*w+1);
				}

				else if(vi.IsYV12())
				{
				
					unsigned char * wpU=work->GetWritePtr(PLANAR_U);
					unsigned char * wpV=work->GetWritePtr(PLANAR_V);
					const int  wpitchUV=work->GetPitch(PLANAR_U);
					const unsigned char * fpU=Frame->GetReadPtr(PLANAR_U);
					const unsigned char * fpV=Frame->GetReadPtr(PLANAR_V);
					const int  fpitchUV=Frame->GetPitch(PLANAR_U);

					for(int h=ty/2;h<by/2;h++)
						for(int w=lx/2;w<rx/2;w++)
						{
						// copy u and v channels
							*(wpU+h*wpitchUV+w)=*(fpU+h*fpitchUV+w);
							*(wpV+h*wpitchUV+w)=*(fpV+h*fpitchUV+w);
						}
				}

				else if(vi.IsRGB())
				{
					for(int h=ty;h<by;h++)
						for(int w=lx;w<rx;w++)
						{
								// copy  green  blue channels .
							*(wp+h*wpitch+kb*w+1)=*(fp+h*fpitch+kb*w+1);
							*(wp+h*wpitch+kb*w+2)=*(fp+h*fpitch+kb*w+2);
						}
				}
			}



	return work;	
	
}

	

double	NeuralNetLN::getoutput()
					{
						
								
						double sum=0.0;
						for ( int in=0; in<inodes; in++)
						{
							sum+= input[in]*wh[in];
						}
						return(sum);	

								
					}


void  NeuralNetLN::getcase(const unsigned char * fp, int fpitch,int kb, int h,int w)
{
	for(int i=0; i< ypts; i++)
		for(int n=0; n<xpts; n++)
			input[i*xpts+n]= *(fp+(h+i)*fpitch+kb*(w+n));
		input[xpts*ypts]=1;	// bias node value
}

void  NeuralNetLN::sumdedwt(double err, double *dwtnew)
{
	for ( int in=0; in<inodes; in++)
	
		dwtnew[in]-= err*input[in];
}



void  NeuralNetLN::adjustWeights(double *dwtold,double *dwtnew)
								 

{
	// resilient propogation

	for(int hn=0; hn<inodes; hn++)
	{
		
		

			if(dwtold[hn] * dwtnew[hn] >0)
			{	// the de/dwt has not changed sign
				
						// use minimum of (deltamax and delta[]*yetaplus)
				delta[hn] = deltamax < delta[hn] * yetaplus ?
										 deltamax : delta[hn] * yetaplus;
						// sign(val) returns -1, 0 or +1 based on val sign)
				deltaw[hn] = - sign(dwtnew[hn])* delta[hn];
				
				wh[hn] += deltaw[hn];
			}

			else if (dwtold[hn] * dwtnew[hn] <0)
			{
				// de/dwt has changed sign
						// use max of ( deltamin and delta[]*yetaminus)
				delta[hn] = deltamin >= delta[hn] * yetaminus ?
										 deltamin : delta[hn] * yetaminus;
				wh[hn] -= deltaw[hn];// go back to prev value as we over stepped a minimum

				dwtnew[hn] =0;	// to ensure that next iter when we come to this function  no update of this weight occurs


			}

			else // zero
			{
				deltaw[hn] = - sign(dwtnew[hn]) * delta[hn];

				wh[hn] += deltaw[hn];
			}
		
	}

}

int NeuralNetLN::sign( double val)
{
	if (val > 0)
		return (1);
	if(val<0)
		return (-1);
	return (0);
}
/***************************************************************/
// This is the function that created the filter, when the filter has been called.
// This can be used for simple parameter checking, so it is possible to create different filters,

AVSValue __cdecl Create_NeuralNetLN(AVSValue args, void* user_data, IScriptEnvironment* env) 
{
	if(!args[0].AsClip()->GetVideoInfo().IsRGB24() && !args[0].AsClip()->GetVideoInfo().IsRGB32() 
		&& !args[0].AsClip()->GetVideoInfo().IsYUY2() && !args[0].AsClip()->GetVideoInfo().IsYV12())
		env->ThrowError("NeuralNetLN: This color format is not supported here");
	int nf=args[0].AsClip()->GetVideoInfo().num_frames;
	if(args[2].AsInt(0)<0 || args[2].AsInt(0)> args[3].AsInt(nf-1))
		env->ThrowError("NeuralNetLN: Frame number must be positive and end frame not less than start frame");
	if(args[2].AsInt(0)>=nf || args[3].AsInt(nf-1)>=nf)
		env->ThrowError("NeuralNetLN:this clip which has only %d frames",nf);
	if(!args[0].AsClip()->GetVideoInfo().pixel_type != !args[1].AsClip()->GetVideoInfo().pixel_type
		|| !args[0].AsClip()->GetVideoInfo().width != !args[1].AsClip()->GetVideoInfo().width
		|| !args[0].AsClip()->GetVideoInfo().height != !args[1].AsClip()->GetVideoInfo().height)
		env->ThrowError("NeuralNetLN:The input clip and trainer clip have differing color formats or dimensions");
	if(args[1].AsClip()->GetVideoInfo().num_frames<=args[18].AsInt(0))
		env->ThrowError("NeuralNetLN:the training clip does not have the specified frame");
	
	int xx=args[0].AsClip()->GetVideoInfo().width-1;
	int yy=args[0].AsClip()->GetVideoInfo().height-1;
	
	
	if(args[4].AsInt(0)<0 || args[5].AsInt(0) <0 || args[8].AsInt(0)<0 || args[9].AsInt(0)<0)
		env->ThrowError("NeuralNetLN: coordinates should not be -ve");
	if(args[6].AsInt(xx)<args[4].AsInt(0)+args[13].AsInt(9) || args[7].AsInt(yy) <args[5].AsInt(0)+args[14].AsInt(9)
		|| args[6].AsInt(xx)>xx || args[7].AsInt(yy)>yy)		
		env->ThrowError("NeuralNetLN: window x and y dimensions should not be less than  xpts or ypts and must be within frame");

	
	if(args[10].AsInt(xx)<args[8].AsInt(args[4].AsInt(0))+args[13].AsInt(9) 
		|| args[11].AsInt(args[7].AsInt(yy)) <args[9].AsInt(args[5].AsInt(0))+args[14].AsInt(9)
		|| args[10].AsInt(args[6].AsInt(xx))>xx || args[11].AsInt(args[7].AsInt(yy))>yy)		
		env->ThrowError("NeuralNetLN: end window x and y dimensions should not be less than  xpts or ypts and must be within frame");
	

	if(args[12].AsInt(9) <1 || args[13].AsInt(9)<1 || args[12].AsInt(9)*args[13].AsInt(9)>121 
		|| args[12].AsInt(9)*args[13].AsInt(9)<3|| (args[12].AsInt(9) & 1) ==0 || (args[13].AsInt(9) & 1) ==0)
		env->ThrowError("NeuralNetLN:xpts and ypts must be odd numbers 1 to 119, and xpts * ypts must be 3 to 121");

	int wx, wy;
	if(args[23].AsBool(true))
	{
		wx=xx-args[12].AsInt(9);
		wy= yy -args[13].AsInt(9);
		if(args[14].AsInt(args[4].AsInt(0)) <0 || args[14].AsInt(args[4].AsInt(0)) > wx
			|| args[16].AsInt(wx) <0 || args[16].AsInt(wx) > wx)
			env->ThrowError("NeuralNetRP: The training line x coordinates must be within frame and less than frame width- xpts");

		if(args[15].AsInt(args[5].AsInt(0)) <0 || args[15].AsInt(args[5].AsInt(0)) > wy
			|| args[17].AsInt(wy) <0 || args[17].AsInt(wy) > wy)
			env->ThrowError("NeuralNetRP: The training line y coordinates must be within frame and less than frame height- ypts");
	}

	else
	{
		wx=xx;
		wy=yy;

		if(args[14].AsInt(args[4].AsInt(0)) <0 || args[14].AsInt(args[4].AsInt(0))>args[16].AsInt(wx)
		|| args[16].AsInt(wx) > wx)
			env->ThrowError("NeuralNetRP: training window x coordinates be +ve, within frame ");

		if(args[15].AsInt(args[5].AsInt(0)) <0 || args[15].AsInt(args[5].AsInt(0))>args[17].AsInt(wy)
		|| args[17].AsInt(wy) > wy)
			env->ThrowError("NeuralNetRP: The  training window y coordinates be +ve, within frame ");
	}
	
	
	if(args[19].AsFloat(0.01) <0 )
		env->ThrowError("NeuralNetLN:the weight factors must be positive numbers");
	if(args[20].AsInt(2380) <=0)
		env->ThrowError("NeuralNetLN:number of iterations must be positive number");
	if(args[21].AsFloat(0.00005) <=0)
		env->ThrowError("NeuralNetLN:minstep must not be 0 or negative");
	
	if(args[22].AsInt(1)<0 || args[22].AsInt(1)>10)
		env->ThrowError("NeuralNetLN:bestof can be between 1 and 10 only");
	if(args[21].AsInt(1)<0 || args[21].AsInt(1)>10)
		env->ThrowError("NeuralNetLN:wset can be between 1 and 10 only");
	
			return  new NeuralNetLN(	args[0].AsClip(),			// clip for input
								args[1].AsClip(),			// training clip
								args[2].AsInt(0),			// start frame number to process						
								args[3].AsInt(nf-1),		// end frame number to process
								
								args[4].AsInt(0),			// rectangle start left x, 
								args[5].AsInt(0),			// rectangle start top y,  
								args[6].AsInt(xx),			// start right x 
								args[7].AsInt(yy),			// start bottom y  
								args[8].AsInt(args[4].AsInt(0)),		// rectangle end left x, 
								args[9].AsInt(args[5].AsInt(0)),		// rectangle end top y,  
								args[10].AsInt(args[6].AsInt(xx)),		// end rightx 
								args[11].AsInt(args[7].AsInt(yy)),		// end bot y
								
							
								args[12].AsInt(3),			// input points in x direction
								args[13].AsInt(3),			// input points in y direction
								args[14].AsInt(args[4].AsInt(0)),	// training clip window left x
								args[15].AsInt(args[5].AsInt(0)),	// training clip window top y
								args[16].AsInt(wx), // training clip window right x
								args[17].AsInt(wy), // training clip window bottom y
								
								args[18].AsInt(0),			// trainer clip frame number to use
								args[19].AsFloat(0.0005),	// input layer weights factor
								
								args[20].AsInt(600),	// number of iterations
								
								args[21].AsInt(1),			// weight set number (start wts are this set)
								
								args[22].AsInt(1),				// use best (least error) weights out of this many different starting weights.
								
								args[23].AsBool(true),			// training along diagonal line or rectangle?
							
								env);
	



	
}
// The following function is the function that actually registers the filter in AviSynth
// It is called automatically, when the plugin is loaded to see which functions this filter contains.
/*
extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
    env->AddFunction("NeuralNetLN", "cc[sf]i[ef]i[lx]i[ty]i[rx]i[by]i[elx]i[ety]i[erx]i[eby]i"
		"[xpts]i[ypts]i[tlx]i[tty]i[trx]i[tby]i[tf]i[wh]f[iter]i"
		"[wset]i[bestof]i[line]b", Create_NeuralNetLN, 0);
					
    return "NeuralNetLN  Plugin";
}
*/
